import atexit
import shutil
import sys
import os
import zipfile
from xml.etree import ElementTree

import m3u8
import requests

from isubrip.constants import DEFAULT_CONFIG_PATH, PACKAGE_NAME, PYPI_RSS_URL, TEMP_FOLDER_PATH
from isubrip.exceptions import ConfigError, DefaultConfigNotFound
from isubrip.namedtuples import MovieData
from isubrip.playlist_downloader import PlaylistDownloader
from isubrip.scraper import Scraper
from isubrip.subtitles import Subtitles
from isubrip.utils import find_config_file, format_title, parse_config


def main() -> None:
    # Check if at least one argument was passed
    if len(sys.argv) < 2:
        print_usage()
        exit(1)

    # Assure default config file exists
    try:
        default_config_path = os.path.join(os.path.dirname(sys.modules[PACKAGE_NAME].__file__), DEFAULT_CONFIG_PATH)

    except KeyError:
        raise DefaultConfigNotFound(f"Default config file could not be found.")

    # Load default and user (if it exists) config files
    config_files = [default_config_path]
    user_config_path = find_config_file()

    if user_config_path is not None:
        config_files.append(find_config_file())

    try:
        config = parse_config(*config_files)

    except (ConfigError, FileNotFoundError) as e:
        raise ConfigError(e)

    # Set `Subtitles` settings from config
    Subtitles.remove_duplicates = config.subtitles["remove-duplicates"]
    Subtitles.fix_rtl = config.subtitles["fix-rtl"]
    Subtitles.rtl_languages = config.subtitles["rtl-languages"]

    download_path: str
    download_to_temp: bool

    # Set download path to temp folder "zip" setting is used
    if config.downloads["zip"]:
        download_path = TEMP_FOLDER_PATH
        download_to_temp = True

    else:
        download_path = config.downloads["folder"]
        download_to_temp = False

    if config.general["check-for-updates"]:
        check_for_updates()

    for idx, url in enumerate(sys.argv[1:]):
        if idx > 0:
            print()  # Print newline between different movies

        print(f"Scraping {url}...")

        try:
            movie_data: MovieData = Scraper.find_movie_data(url, config.scraping["user-agent"])

        except Exception as e:
            print(f"Error: {e}")
            continue

        print(f"Found movie: {movie_data.name}")

        if movie_data.playlist is None:
            print(f"Error: No valid playlist could be found.")
            print("\n--------------------------------------------------")
            continue

        m3u8_playlist: m3u8.M3U8 = m3u8.load(movie_data.playlist)

        # Create temp folder
        if download_to_temp:
            current_download_path = os.path.join(download_path, f"{format_title(movie_data.name)}.iT.WEB")
            os.makedirs(current_download_path, exist_ok=True)
            atexit.register(shutil.rmtree, current_download_path)

        else:
            current_download_path = download_path

        downloaded_subtitles_list = []
        subtitles_count = 0

        with PlaylistDownloader(config.downloads["user-agent"]) as playlist_downloader:
            for subtitles in Scraper.find_subtitles(m3u8_playlist, config.downloads["languages"]):
                subtitles_count += 1
                print(f"Downloading \"{subtitles.language_name}\" ({subtitles.language_code}) subtitles...")
                downloaded_subtitles = playlist_downloader.download_subtitles_file(movie_data, subtitles, current_download_path, config.downloads["format"])

                # Assure subtitles downloaded successfully
                if os.path.isfile(downloaded_subtitles):
                    downloaded_subtitles_list.append(downloaded_subtitles)

            if download_to_temp:
                if len(downloaded_subtitles_list) == 1:
                    shutil.copy(downloaded_subtitles_list[0], config.downloads["folder"])

                elif len(downloaded_subtitles_list) > 1:
                    # Create zip archive
                    print(f"Creating zip archive...")
                    archive_name = f"{format_title(movie_data.name)}.iT.WEB.zip"
                    archive_path = os.path.join(current_download_path, archive_name)

                    zf = zipfile.ZipFile(archive_path, compression=zipfile.ZIP_DEFLATED, mode='w')

                    for file in downloaded_subtitles_list:
                        zf.write(file, os.path.basename(file))

                    zf.close()
                    shutil.copy(archive_path, config.downloads["folder"])

                # Remove current temp dir
                shutil.rmtree(current_download_path)
                atexit.unregister(shutil.rmtree)

        print(f"\n{len(downloaded_subtitles_list)}/{subtitles_count} matching subtitles for \"{movie_data.name}\" were downloaded to \"{os.path.abspath(config.downloads['folder'])}\".")

        if idx < (len(sys.argv) - 2):
            print("\n--------------------------------------------------")


def check_for_updates() -> None:
    """Check and print if a newer version of the package is available."""
    # If anything breaks, just skip update check
    try:
        current_version = sys.modules[PACKAGE_NAME].__version__

        response = requests.get(PYPI_RSS_URL).text
        xml_data = ElementTree.fromstring(response)
        latest_version = xml_data.find("channel/item/title").text

        # If the latest PyPI release is different from current one, print a message
        if latest_version != current_version:
            print(f"Note: You are currently using version {current_version} of {PACKAGE_NAME}, however version {latest_version} is available.",
                  f"\nConsider upgrading by running \"python3 -m pip install --upgrade {PACKAGE_NAME}\"\n")

    except Exception:
        return


def print_usage() -> None:
    """Print usage information."""
    print(f"Usage: {PACKAGE_NAME} <iTunes movie URL> [iTunes movie URL...]")


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from datetime import time
from typing import Union

from isubrip.constants import RTL_CHAR, RTL_CONTROL_CHARS, TIMESTAMP_REGEX
from isubrip.enums import SubtitlesFormat


class Subtitles:
    """An object representing subtitles, made out of paragraphs."""

    remove_duplicates = False
    fix_rtl = False
    rtl_languages = []

    def __init__(self, language_code: str = None):
        """Initalize a new Subtitles object."""
        self.language_code = language_code
        self.paragraphs = []

    def __add__(self, paragraph: Paragraph) -> Subtitles:
        """
        Add a new paragraph to current subtitles.

        Args:
            paragraph (Paragraph): A paragraph object to append.
        """
        self.add_paragraph(paragraph)
        return self

    def _dumps_vtt(self) -> str:
        """
        Dump subtitles to a string in VTT format.

        Returns:
            str: The subtitles formatted as a string in VTT format.
        """
        subtitles_str = "WEBVTT\n\n"

        for paragraph in self.paragraphs:
            subtitles_str += f"{paragraph.to_string(SubtitlesFormat.VTT)}\n\n"

        return subtitles_str.rstrip('\n')

    def _dumps_srt(self) -> str:
        """
        Dump subtitles to a string in SRT format.

        Returns:
            str: The subtitles formatted as a string in SRT format.
        """
        subtitles_str = ""

        for idx, paragraph in enumerate(self.paragraphs):
            subtitles_str += f"{(idx + 1)}\n{paragraph.to_string(SubtitlesFormat.SRT)}\n\n"

        return subtitles_str.rstrip('\n')

    @staticmethod
    def _split_timestamp(timestamp: str) -> tuple[time, time]:
        """
        Splits a timestamp into start and end.

        Args:
            timestamp (str): A subtitles timestamp. For example: "00:00:00.000 --> 00:00:00.000"

        Returns:
            tuple(time, time): A tuple containing start and end times as a datetime object.
        """
        # Support ',' character in timestamp's milliseconds (used in srt format).
        timestamp = timestamp.replace(',', '.')

        start_time, end_time = timestamp.split(" --> ")
        return time.fromisoformat(start_time), time.fromisoformat(end_time)

    def add_paragraph(self, paragraph: Paragraph) -> None:
        """
        Add a new paragraph to current subtitles.

        Args:
            paragraph (Paragraph): A paragraph object to append.
        """
        # Don't append if `remove-duplicates` is set to true and paragraph is same as previous one
        if not (Subtitles.remove_duplicates and len(self.paragraphs) > 0 and self.paragraphs[-1] == paragraph):
            # Fix RTL before appending if `fix-rtl` is set to true and language is an RTL language
            if Subtitles.fix_rtl and self.language_code in Subtitles.rtl_languages:
                paragraph.fix_rtl()

            self.paragraphs.append(paragraph)

    def append_subtitles(self, subtitles: Subtitles) -> None:
        """
        Append an existing subtitles object.

        Args:
            subtitles (Subtitles): Subtitles object to append to current subtitels.
        """
        for paragraph in subtitles.paragraphs:
            self.add_paragraph(paragraph)

    @staticmethod
    def loads(subtitles_data: str) -> Subtitles:
        """
        Load subtitles from a string.

        Args:
            subtitles_data (str): Subtitles data to load.

        Returns:
            Subtitles: A Subtitles object loaded from the string.
        """
        subtitles_obj = Subtitles()

        regex_split = re.split(rf"^(?:[0-9]+\n)?({TIMESTAMP_REGEX}).*\n", subtitles_data, flags=re.MULTILINE)

        paragraph_timestamp: Union[str, None] = None
        paragraph_text = ""

        for line in regex_split:
            if re.match(TIMESTAMP_REGEX, line):
                if paragraph_timestamp is not None:
                    timestamps = Subtitles._split_timestamp(paragraph_timestamp)
                    subtitles_obj += Paragraph(timestamps[0], timestamps[1], paragraph_text.rstrip("\n"))
                paragraph_timestamp = line
                paragraph_text = ""

            elif paragraph_timestamp is not None:
                paragraph_text += line

        if paragraph_timestamp is not None:
            timestamps = Subtitles._split_timestamp(paragraph_timestamp)
            subtitles_obj.add_paragraph(Paragraph(timestamps[0], timestamps[1], paragraph_text.rstrip("\n")))

        return subtitles_obj

    def dumps(self, subtitles_format: SubtitlesFormat = SubtitlesFormat.VTT) -> str:
        """
        Dump subtitles to a string.

        Args:
            subtitles_format (SubtitlesFormat): Subtitles format specification to use.

        Returns:
            str: The subtitles formatted as a string matching the specified subtitles format.
        """
        if subtitles_format == SubtitlesFormat.VTT:
            return self._dumps_vtt()

        elif subtitles_format == SubtitlesFormat.SRT:
            return self._dumps_srt()


class Paragraph:
    """An object represnting a subtitles paragraph."""

    def __init__(self, start_time: time, end_time: time, text: str):
        """
        Create a new Paragraph object.

        Args:
            start_time (time): Paragraph start time.
            end_time (time): Paragraph end time.
            text: Paragraph text.
        """
        self.start_time = start_time
        self.end_time = end_time
        self.text = text

    def __eq__(self, other) -> bool:
        """
        Check whether two Paragraph objects are equal.

        Args:
            other: A Paragraph object to compare to.

        Returns:
            bool: True if Paragraph objects are equal, False otherwise.
        """
        return isinstance(other, Paragraph) and \
            self.start_time == other.start_time and self.end_time == other.end_time and self.text == other.text

    def fix_rtl(self) -> None:
        """Fix paragraph direction to RTL."""
        # Remove previous RTL-related formatting
        for char in RTL_CONTROL_CHARS:
            self.text = self.text.replace(char, "")

        # Add RLM char at the start and on every new line
        self.text = RTL_CHAR + self.text.replace("\n", f"\n{RTL_CHAR}")

    def to_string(self, subtitles_format: SubtitlesFormat) -> str:
        """
        Convert current paragraph to a subtitles-formatted string.

        Args:
            subtitles_format (SubtitlesFormat): Subtitles format specification to use.

        Returns:
            str: The paragraph formatted as a string matching the specified subtitles format.
        """
        time_format = None

        if subtitles_format == SubtitlesFormat.VTT:
            time_format = "%H:%M:%S.%f"

        elif subtitles_format == SubtitlesFormat.SRT:
            time_format = "%H:%M:%S,%f"

        timestamp = f"{self.start_time.strftime(time_format)[:-3]} --> {self.end_time.strftime(time_format)[:-3]}"
        return f"{timestamp}\n{self.text}"



# --------------- Scraping --------------- #

class ScrapeError(Exception):
    """An issue while scraping the page."""
    pass


class InvalidURL(ScrapeError):
    """An invalid URL has been used."""
    pass


class PageLoadError(ScrapeError):
    """The Page did not load properly."""
    pass


class PlaylistDownloadError(ScrapeError):
    """The playlist could not be downloaded."""
    pass


# --------------- Config --------------- #

class ConfigError(Exception):
    """An issue with a config file."""
    pass


class DefaultConfigNotFound(ConfigError):
    """Default config file could not be found."""
    pass


class ConfigValueMissing(ConfigError):
    """A required config value is missing."""
    pass


class InvalidConfigValue(ConfigError):
    """An invalid value has been used."""
    pass



import os
import sys
from typing import Union

from isubrip.config import Config
from isubrip.constants import CONFIG_FILE_NAME, APPDATA_FOLDER_NAME, APPDATA_PATH_LINUX, APPDATA_PATH_MACOS, APPDATA_PATH_WINDOWS


def find_appdata_path() -> str:
    """Return the path to appdata folder.

    Returns:
        Union[str, None]: A string with the appdata folder.
    """
    # Windows
    if sys.platform == "win32":
        return APPDATA_PATH_WINDOWS

    # Linux
    elif sys.platform == "linux":
        return APPDATA_PATH_LINUX

    # MacOS
    elif sys.platform == "darwin":
        return APPDATA_PATH_MACOS


def find_config_file() -> Union[str, None]:
    """Return the path to user's config file (if it exists).

    Returns:
        Union[str, None]: A string with the path to user's config file if it's found, and None otherwise.
    """
    config_path = os.path.join(find_appdata_path(), APPDATA_FOLDER_NAME, CONFIG_FILE_NAME)

    if (config_path is not None) and (os.path.exists(config_path)):
        return config_path

    return None


def format_title(title: str) -> str:
    """Format movie title to a standardized title that can be used as a file name.

    Args:
        title (str): An iTunes movie title.

    Returns:
        str: The title, in a file-name-friendly format.
    """
    # Replacements will be done in the same order of this list
    replacement_pairs = [
        (': ', '.'),
        (' - ', '-'),
        (', ', '.'),
        ('. ', '.'),
        (' ', '.'),
        ('(', ''),
        (')', '')
    ]

    for pair in replacement_pairs:
        title = title.replace(pair[0], pair[1])

    return title


def parse_config(file_path: str, *file_paths: str) -> Config:
    """
    Parse config files by order and return a Config object.

    Args:
        file_path (str): A config file to parse.
        *file_paths (str, optional): Additional config files to parse (will override previous settings).

    Returns:
        Config: A parsed Config object.
    """
    config = Config()
    file_paths: tuple = (file_path,) + file_paths

    for file_path in file_paths:
        with open(file_path, 'r') as data:
            config.loads(data.read())

    return config



from typing import NamedTuple, Type, Union

from isubrip.enums import SubtitlesType


class MovieData(NamedTuple):
    """A named tuple containing a movie name, id, and M3U8 playlist."""
    id: str
    name: str
    release_year: int
    playlist: Union[str, None]


class SubtitlesData(NamedTuple):
    """A named tuple containing language code, language name, type and playlist URL for subtitles."""
    language_code: str
    language_name: str
    subtitles_type: SubtitlesType
    playlist_url: str


class ConfigSetting(NamedTuple):
    category: str
    key: str
    types: Union[tuple, Type]



import os

from tempfile import gettempdir
from xdg import xdg_config_home

# General
PACKAGE_NAME = "isubrip"
PYPI_RSS_URL = "https://pypi.org/rss/project/isubrip/releases.xml"  # Used for checking updates

# Paths
DEFAULT_CONFIG_PATH = os.path.join("resources", "default_config.toml")
APPDATA_PATH_WINDOWS = f"{os.getenv('appdata')}"
APPDATA_PATH_LINUX = f"{xdg_config_home().resolve()}"
APPDATA_PATH_MACOS = r"~/Library/Application Support"
APPDATA_FOLDER_NAME = 'iSubRip'
CONFIG_FILE_NAME = "config.toml"
TEMP_FOLDER_PATH = os.path.join(gettempdir(), 'iSubRip')

# Lists
VALID_ARCHIVE_FORMATS = ["zip", "tar", "gztar"]

# RegEx
ITUNES_STORE_REGEX = r"https?://itunes.apple.com/[a-z]{2}/movie/[a-zA-Z0-9\-%]+/id[0-9]+($|(\?.*))"
TIMESTAMP_REGEX = r"^[0-9]{2}:[0-9]{2}:[0-9]{2}[\.,][0-9]{3} --> [0-9]{2}:[0-9]{2}:[0-9]{2}[\.,][0-9]{3}"

# Unicode
RTL_CONTROL_CHARS = ('\u200e', '\u200f', '\u202a', '\u202b', '\u202c', '\u202d', '\u202e')
RTL_CHAR = '\u202b'



import json
import re
from datetime import datetime
from typing import Union, Iterator
from urllib.error import HTTPError

import m3u8
from bs4 import BeautifulSoup
from bs4.element import NavigableString, Tag
from m3u8.model import M3U8
from requests.sessions import session

from isubrip.enums import SubtitlesType
from isubrip.constants import ITUNES_STORE_REGEX
from isubrip.namedtuples import MovieData, SubtitlesData
from isubrip.exceptions import InvalidURL, PageLoadError


class Scraper:
    """A class for scraping and downloading subtitles off of iTunes movie pages."""

    @staticmethod
    def find_movie_data(itunes_url: str, user_agent: Union[str, None] = None) -> MovieData:
        """
        Scrape an iTunes store page to find movie info and it's M3U8 playlist.

        Args:
            itunes_url (str): An iTunes store movie URL.
            user_agent (str | None, optional): A dictionary with iTunes data loaded from a JSON response. Defaults to None.
        
        Raises:
            InvalidURL: `itunes_url` is not a valid iTunes store movie URL.
            PageLoadError: HTML page did not load properly.  

        Returns:
            MovieData: A MovieData (NamedTuple) object with movie's name, and an M3U8 object of the playlist
            if the playlist is found. None otherwise.
        """
        # Check whether URL is valid
        if re.match(ITUNES_STORE_REGEX, itunes_url) is None:
            raise InvalidURL(f"{itunes_url} is not a valid iTunes movie URL.")

        user_agent_header = {"User-Agent": user_agent} if user_agent is not None else None
        response = session().get(itunes_url, headers=user_agent_header)

        # Response is JSON formatted
        if "application/json" in response.headers['content-type']:
            try:
                json_data = json.loads(response.text)

            except json.JSONDecodeError:
                raise PageLoadError("Recieved an invalid JSON response.")

            return Scraper._find_playlist_data_json_(json_data)

        # Response is HTML formatted
        else:
            html_data = BeautifulSoup(response.text, "lxml")
            return Scraper._find_playlist_data_html_(html_data)

    @staticmethod
    def _find_playlist_data_json_(json_data: dict) -> MovieData:
        """
        Scrape an iTunes JSON response to find movie info and it's M3U8 playlist.

        Args:
            json_data (dict): A dictionary with iTunes data loaded from a JSON response.

        Returns:
            MovieData: A MovieData (NamedTuple) object with movie's name, and an M3U8 object of the playlist
            if the playlist is found. None otherwise.
        """
        movie_id = json_data["pageData"]["id"]
        movie_data = json_data["storePlatformData"]["product-dv"]["results"][movie_id]

        movie_title = movie_data["nameRaw"]
        movie_release_year = datetime.strptime(movie_data["releaseDate"], '%Y-%m-%d').year

        # Loop safely to find a matching playlist
        for offer in movie_data["offers"]:
            if isinstance(offer.get("type"), str) and offer["type"] in ["buy", "rent"]:
                if isinstance(offer.get("assets"), list) and len(offer["assets"]) > 0:
                    for asset in offer["assets"]:
                        m3u8_url: str = asset["hlsUrl"]

                        # Assure playlist is valid
                        try:
                            m3u8.load(m3u8_url)

                        # If m3u8 playlist is invalid, skip it
                        except ValueError:
                            continue

                        except HTTPError:
                            continue

                        return MovieData(movie_id, movie_title, movie_release_year, m3u8_url)

        return MovieData(movie_id, movie_title, movie_release_year, None)

    @staticmethod
    def _find_playlist_data_html_(html_data: BeautifulSoup) -> MovieData:
        """
        Scrape an iTunes HTML page to find movie info and it's M3U8 playlist.

        Args:
            html_data (BeautifulSoup): A BeautifulSoup object of the page.

        Raises:
            PageLoadError: HTML page did not load properly.

        Returns:
            MovieData: A MovieData (NamedTuple) object with movie's name, and an M3U8 object of the playlist
            if the playlist is found. None otherwise.
        """
        # NOTE: This function is less reliable than `_find_playlist_data_json_`.

        movie_id_tag: Union[Tag, NavigableString, None] = html_data.find("meta", attrs={"name": "apple:content_id"})
        if not isinstance(movie_id_tag, Tag):
            raise PageLoadError("HTML page did not load properly.")

        movie_id: str = movie_id_tag.attrs["content"]

        # Scrape a dictionary on the webpage for playlists data
        shoebox_data_tag: Union[Tag, NavigableString, None] = html_data.find("script", attrs={"id": "shoebox-ember-data-store", "type": "fastboot/shoebox"})

        # fastboot/shoebox data could not be found
        if not isinstance(shoebox_data_tag, Tag):
            raise PageLoadError("fastboot/shoebox data could not be found.")

        # Convert to dictionary structure
        shoebox_data: dict[str, dict] = json.loads(str(shoebox_data_tag.contents[0]).strip())

        # Loop safely to find a matching playlist
        if isinstance(shoebox_data[movie_id].get("included"), list):
            movie_data: dict = shoebox_data[movie_id]
            movie_title: str = movie_data["data"]["attributes"]["name"]
            movie_release_year = datetime.strptime(movie_data["data"]["attributes"]["releaseDate"], '%Y-%m-%d').year

            for item in movie_data["included"]:
                if isinstance(item.get("type"), str) and item["type"] == "offer":
                    if isinstance(item.get("attributes"), dict) and \
                        isinstance(item["attributes"].get("assets"), list) and \
                            len(item["attributes"]["assets"]) > 0:

                        for asset in item["attributes"]["assets"]:
                            if isinstance(asset, dict) and isinstance(asset.get("hlsUrl"), str):
                                m3u8_url: str = item["attributes"]["assets"][0]["hlsUrl"]

                                # Try to load the playlist, to assure it's valid
                                try:
                                    m3u8.load(m3u8_url)

                                # If m3u8 playlist is invalid, skip it
                                except (ValueError, HTTPError):
                                    continue

                                return MovieData(movie_id, movie_title, movie_release_year, m3u8_url)
        else:
            raise PageLoadError("Invalid shoebox data.")

        return MovieData(movie_id, movie_title, movie_release_year, None)

    @staticmethod
    def find_subtitles(main_playlist: M3U8, subtitles_filter: Union[list, None] = None) -> Iterator[SubtitlesData]:
        """
        Find and yield playlists within main_playlist for subtitles that match a filter.

        Args:
            main_playlist (M3U8): an M3U8 object of the main playlist.
            subtitles_filter (list, optional): A list of subtitles language codes (ISO 639-1) or names to use as a filter. Defaults to None.

        Yields:
            SubtitlesData: A NamedTuple with a matching playlist, and it's metadata:
            Language Code, Language Name, SubtitlesType, Playlist URL.
        """
        if subtitles_filter is not None:
            # Convert filters to lower-case for case-insensitive matching
            subtitles_filter = [f.lower() for f in subtitles_filter]

        for playlist in main_playlist.media:
            # Check whether playlist is valid and matches filter
            # "group_id" can be either ["subtitles_ak" / "subtitles_vod-ak-amt.tv.apple.com"] or
            # ["subtitles_ap2" / "subtitles_ap3" / "subtitles_vod-ap-amt.tv.apple.com" / "subtitles_vod-ap1-amt.tv.apple.com" / "subtitles_vod-ap3-amt.tv.apple.com"]
            if (playlist.type == "SUBTITLES") and (playlist.group_id in ("subtitles_ak", "subtitles_vod-ak-amt.tv.apple.com")):

                language_code: str = playlist.language
                language_name: str = playlist.name
                sub_type: SubtitlesType = SubtitlesType.NORMAL

                # Playlist does not match filter
                if subtitles_filter is not None and not (language_code.lower() in subtitles_filter or language_name in subtitles_filter):
                    continue

                # Change subtitles type to "Forced" / "Closed Captions" if needed.
                if playlist.forced == "YES":
                    sub_type = SubtitlesType.FORCED

                elif playlist.characteristics is not None and "public.accessibility" in playlist.characteristics:
                    sub_type = SubtitlesType.CC

                yield SubtitlesData(language_code, language_name, sub_type, playlist.uri)



import os
from setuptools import setup


def get_version(relative_path: str):
    current_path = os.path.abspath(os.path.dirname(__file__))

    with open(os.path.join(current_path, relative_path), 'r') as fp:
        file_data: str = fp.read()

    for line in file_data.splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise RuntimeError("Unable to find version string.")


PACKAGE_NAME: str = "isubrip"

with open("README.md", "r") as file:
    long_description = file.read()


setup(
    name=PACKAGE_NAME,
    version=get_version(f"{PACKAGE_NAME}/__init__.py"),
    author="Michael Yochpaz",
    license="MIT",
    license_files=('LICENSE',),
    description="A Python package for scraping and downloading subtitles from iTunes movie pages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MichaelYochpaz/iSubRip",

    project_urls={
        "Bug Reports": "https://github.com/MichaelYochpaz/iSubRip/issues",
        "Source": "https://github.com/MichaelYochpaz/iSubRip"
    },

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
    ],

    keywords=["iTunes", "movies", "subtitles", "scrape", "scraper", "download", "m3u8"],
    packages=["isubrip"],
    install_requires=["aiohttp", "beautifulsoup4", "lxml", "m3u8", "mergedeep", "requests", "tomli", "xdg"],
    package_data={"isubrip": ["resources/*"]},
    python_requires=">=3.7",
    entry_points={
        "console_scripts":
            ["isubrip = isubrip.__main__:main"]
    },
)



from typing import Any

import tomli
from mergedeep import merge

from isubrip.enums import SubtitlesFormat
from isubrip.exceptions import ConfigValueMissing, InvalidConfigValue
from isubrip.namedtuples import ConfigSetting


class Config:
    """A class for managing iSubRip config files."""

    def __init__(self) -> None:
        """Create a new ConfigManager instance."""
        self.config_dict: dict = {}
        self.config_loaded: bool = False

        # List of valid subtitle formats as strings
        self._valid_subtitle_formats: set = set(item.name.upper() for item in SubtitlesFormat)

    def __getattr__(self, key: str) -> Any:
        """Allow access to config settings using attributes.

        Args:
            key (str): Config key to get.

        Returns:
            Any: The corresponding value for the key in the config.
        """
        if key in self.config_dict:
            return self.config_dict[key]

    def loads(self, config_data: str) -> None:
        """Parse a tomli iSubRip config from a string.

        Args:
            config_data (str): Config file data as a string.

        Raises:
            FileNotFoundError: Config file could not be found in the specified path.
            TOMLDecodeError: Config file is not a valid TOML file.
            ConfigValueMissing: A required config value is missing.
            InvalidConfigValue: An invalid value was used in the config file.

        Returns:
            dict: A dictionary containing all settings.
        """

        # Load settings from default config file
        loaded_data: dict = tomli.loads(config_data)

        if not self.config_loaded:
            temp_config: dict = loaded_data
            self.config_loaded = True

        else:
            temp_config: dict = dict(merge(self.config_dict, loaded_data))

        # Convert download format from string to SubtitlesFormat enum
        if isinstance(temp_config["downloads"]["format"], str) and temp_config["downloads"]["format"].upper() in self._valid_subtitle_formats:
            temp_config["downloads"]["format"] = SubtitlesFormat[temp_config["downloads"]["format"].upper()]

        elif not isinstance(temp_config["downloads"]["format"], SubtitlesFormat):
            raise InvalidConfigValue(f"Invalid config value for downloads.format: {temp_config['downloads']['format']}")

        self._standardize_config_(temp_config)
        self.check_config(temp_config)
        self.config_dict = temp_config

    @staticmethod
    def _standardize_config_(config_dict: dict) -> None:
        """Standardize a config dictionary and fix issues.

        Args:
            config_dict (dict): Config dictionary to standardize.
        """
        # If languages list is empty, change it to None
        if not config_dict["downloads"]["languages"]:
            config_dict["downloads"]["languages"] = None

        # Remove a trialing slash / backslash from path
        if isinstance(config_dict["downloads"]["format"], str):
            config_dict["downloads"]["folder"] = config_dict["downloads"]["folder"].rstrip(r"\/")

    @staticmethod
    def check_config(config_dict: dict) -> None:
        """Check the config for invalid values.
        Raises an error if an invalid value is found.
    
        Args:
            config_dict (dict): Config dictionary to check.

        Raises:
            ConfigValueMissing: A required config value is missing.
            InvalidConfigValue: An invalid value was used in the config file.
        """
        
        # List of config values and their corresponding types
        setting_list = [
            ConfigSetting("general", "check-for-updates", bool),
            ConfigSetting("downloads", "folder", str),
            ConfigSetting("downloads", "format", SubtitlesFormat),
            ConfigSetting("downloads", "languages", (list, type(None))),
            ConfigSetting("downloads", "user-agent", str),
            ConfigSetting("downloads", "zip", bool),
            ConfigSetting("scraping", "user-agent", str),
            ConfigSetting("subtitles", "remove-duplicates", bool),
            ConfigSetting("subtitles", "fix-rtl", bool),
            ConfigSetting("subtitles", "rtl-languages", list),
            ]

        # Assure each config value exists and is of the correct type
        for setting in setting_list:
            if setting.category not in config_dict:
                raise ConfigValueMissing(f"Config category \'{setting.category}\' with required settings is missing.")

            if setting.key in config_dict[setting.category]:
                setting_value = config_dict[setting.category][setting.key]

                if not isinstance(setting_value, setting.types):
                    raise InvalidConfigValue(f"Invalid config value type for {setting.category}.{setting.key}: \'{setting_value}\'\
                    \nExpected {setting.types}, received: {type(setting_value)}.")

            else:
                raise ConfigValueMissing(f"Missing required config value: \'{setting.category}.{setting.key}\'")



from enum import Enum


class SubtitlesType(Enum):
    """Subtitles type (Normal / CC / Forced)."""
    NORMAL = 1
    CC = 2
    FORCED = 3


class SubtitlesFormat(Enum):
    """Subtitles format (srt / vtt)."""
    SRT = 1
    VTT = 2


class ArchiveFormat(Enum):
    """Archive Format (zip / tar / tar.gz)."""
    ZIP = 1
    TAR = 2
    TAR_GZ = 3



import asyncio
import os

import aiohttp
import m3u8

from isubrip.enums import SubtitlesFormat
from isubrip.namedtuples import SubtitlesData, SubtitlesType, MovieData
from isubrip.subtitles import Subtitles
from isubrip.utils import format_title


class PlaylistDownloader:
    """A class for downloading & converting m3u8 playlists into subtitles."""
    def __init__(self, user_agent: str = None) -> None:
        """
        Create a new PlaylistDownloader instance.

        Args:
            user_agent (str): User agent to use when downloading. Uses default user-agent if not set.
        """
        self.session = aiohttp.ClientSession()

        if user_agent is not None:
            self.session.headers.update({"user-agent": user_agent})

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    async def _download_segment(self, segment_url: str) -> str:
        """
        Download an m3u8 segment.

        Args:
            segment_url (str): Segment URL to download.

        Returns:
            str: Downloaded segment data as a string.
        """
        data = await self.session.get(segment_url)
        content = await data.read()
        return content.decode('utf-8')

    @staticmethod
    def _format_file_name(movie_title: str, release_year: int, language_code: str, subtitles_type: SubtitlesType, file_format: SubtitlesFormat) -> str:
        """Generate file name for a subtitles file.

        Args:
            movie_title (str): Movie title.
            release_year(int): Movie release year.
            language_code (str): Subtitles language code.
            subtitles_type (SubtitlesType): Subtitles type.

        Returns:
            str: A formatted file name (does not include a file extension).
        """
        # Add release year only if it's not already included in the title
        movie_release_year_str = '.' + str(release_year) if str(release_year) not in movie_title else ''
        file_name = f"{format_title(movie_title)}{movie_release_year_str}.iT.WEB.{language_code}"

        # Add subtitles type to file name if it's not `NORMAL` (ex: `FORCED`)
        if subtitles_type is not SubtitlesType.NORMAL:
            file_name += f".{subtitles_type.name.lower()}"

        # Add file format to file name (ex: ".vtt")
        file_name += f".{file_format.name.lower()}"
        return file_name

    def close(self) -> None:
        """Close aiohttp session."""
        async_loop = asyncio.get_event_loop()
        close_task = async_loop.create_task(self.session.close())
        async_loop.run_until_complete(asyncio.gather(close_task))

    def download_subtitles(self, subtitles_data: SubtitlesData) -> Subtitles:
        """
        Get subtitles playlist.

        Args:
            subtitles_data (SubtitlesData): A SubtitlesData namedtuple with information about the subtitles.

        Returns:
            Subtitles: A Subtitles object representing the downloaded subtitles.
        """
        subtitles = Subtitles(subtitles_data.language_code)
        playlist = m3u8.load(subtitles_data.playlist_url)

        async_loop = asyncio.get_event_loop()
        async_tasks = [async_loop.create_task(self._download_segment(segment.absolute_uri)) for segment in playlist.segments]
        segments = async_loop.run_until_complete(asyncio.gather(*async_tasks))

        for segment in segments:
            subtitles.append_subtitles(Subtitles.loads(segment))

        return subtitles

    def download_subtitles_file(self, movie_data: MovieData, subtitles_data: SubtitlesData, output_dir: str, file_format: SubtitlesFormat = SubtitlesFormat.VTT) -> str:
        """
        Download subtitles playlist to a file.

        Args:
            movie_data (MovieData): A MovieData namedtuple with information about the movie.
            subtitles_data (SubtitlesData): A SubtitlesData namedtuple with information about the subtitles.
            output_dir (str): Path to output directory (where the file will be saved).
            file_format (SubtitlesFormat, optional): File format to use for the downloaded file. Defaults to `VTT`.

        Returns:
            str: Path to the downloaded subtitles file.
        """
        file_name = self._format_file_name(
            movie_data.name,
            movie_data.release_year,
            subtitles_data.language_code,
            subtitles_data.subtitles_type,
            file_format)
        path = os.path.join(output_dir, file_name)

        with open(path, 'w', encoding="utf-8") as f:
            f.write(self.download_subtitles(subtitles_data).dumps(file_format))

        return path
